/***************************************************************************
 *   Copyright (C) 2005 by Matthias Jauernig, Michael Lahl                 *
 *   rtc@linux-related.de, DJ-Mayday@web.de                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

/** \file simpleNet.cpp
 *  \brief Diese Datei definiert die Neuronale-Netz-Klasse simpleNet.
 *
 * Diese Datei dient zur Definition von simpleNet, einer Klasse, welche
 * Funktionalitt zum Anlernen und Durchlaufen eines Neuronalen Netzes
 * bereitstellt.
 * $Id: simpleNet.cpp,v 1.10 2005/05/30 04:36:50 mjauerni Exp $
 */
#include <cstdlib>
#include <ctime>
#include <cmath>
#include <string>
#include <fstream>
#include "simpleNet.h"

using std::string;
using std::ios_base;
using std::endl;
using std::ifstream;
using std::ofstream;

/* --- Definition des Konstruktors fr die file_exception ---------------- */
ANN_ERROR::file_exception::file_exception(const string& q){ p = q; }

/* --- Erzeugung des Netzes durch bergabe der gewnschten Dimensionen --- */
simpleNet::simpleNet(const size_t dimI, const size_t dimH, const size_t dimO){
    MemAlloc(dimI,dimH,dimO);
    // Gewichtsmatrizen mit Zufallswerten im Bereich [-0.3...0.3] fllen
    srand(time(NULL));
    for(size_t i=0; i<=m_amountI; i++)
        for(size_t j=0; j<m_amountH; j++)
            m_pIHweights[i][j] = (float)rand()/RAND_MAX*0.6-0.3;
    for(size_t i=0; i<=m_amountH; i++)
        for(size_t j=0; j<m_amountO; j++)
            m_pHOweights[i][j] = (float)rand()/RAND_MAX*0.6-0.3;
}

/* --- Erzeugung des Netzes durch bergabe der Dims und Gewichtsmatrizen - */
simpleNet::simpleNet(const size_t dimI, const size_t dimH, const size_t dimO,
  const float **ihW, const float **hoW){
    MemAlloc(dimI,dimH,dimO);
    // Gewichtsmatrizen mit den bergebenen Werten erzeugen
    for(size_t i=0; i<=m_amountI; i++)
        for(size_t j=0; j<=m_amountH; j++)
            m_pIHweights[i][j] = ihW[i][j];
    for(size_t i=0; i<=m_amountH; i++)
        for(size_t j=0; j<=m_amountO; j++)
            m_pHOweights[i][j] = hoW[i][j];   
}

/* --- Netz erzeugen durch Einlesen von einer Datei ---------------------- */
simpleNet::simpleNet(const string& filename) throw(ANN_ERROR::file_exception){
    size_t dimI, dimH, dimO;
    ifstream in(filename.c_str());
    if(!in)
        throw ANN_ERROR::file_exception
            ("Fehler beim ffnen der Datei zum Laden des Neuronalen Netzes.");
    // Ausnahme werfen bei Einlesefehlern
    in.exceptions(ios_base::failbit|ios_base::badbit|ios_base::eofbit);	
    // Netz-Dimensionen einlesen
    try{
    	in >> dimI;
    	in >> dimH;
    	in >> dimO;
   	}
   	catch(ios_base::failure){
   	    in.close();
        throw ANN_ERROR::file_exception
            ("Fehler beim Einlesen der Dimensionen fr das Neuronale Netz.");
    }
    MemAlloc(dimI,dimH,dimO);
    // Gewichtsmatrizen einlesen
    for(size_t i=0; i<=dimI; i++)
    	for(size_t j=0; j<dimH; j++){
    	    try{
    	        in >> m_pIHweights[i][j];
	        }    
	        catch(ios_base::failure){
	            in.close();
                throw ANN_ERROR::file_exception("Fehler beim Einlesen der 1. Gewichtsmatrix.");
            }
        }
  	for(size_t i=0; i<=dimH; i++)
    	for(size_t j=0; j<dimO; j++){
    		try{
      			in >> m_pHOweights[i][j];
			}
			catch(ios_base::failure){
                in.close();
			    throw ANN_ERROR::file_exception("Fehler beim Einlesen der 2. Gewichtsmatrix.");
			}    
   		}
     in.close();	
}

/* --- Beim Lschen des Neuronalen Netzes allokierten Speicher freigeben - */
simpleNet::~simpleNet(){
	delete[] m_pPrefO; delete[] m_pInI;
    delete[] m_pInH; delete[] m_pInO; delete[] m_pOutI; delete[] m_pOutH; delete[] m_pOutO;
    delete[] m_pErrH; delete[] m_pErrO; delete[] m_pDiffO;
    
    for(size_t i=0; i<=m_amountI; i++)
    	delete m_pIHweights[i];
    for(size_t i=0; i<=m_amountH; i++)
        delete[] m_pHOweights[i];
    for(size_t i=0; i<=m_amountI; i++)
        delete[] m_pIHwDelta[i];
    for(size_t i=0; i<=m_amountH; i++)
        delete m_pHOwDelta[i];
    delete[] m_pIHweights; delete[] m_pHOweights;
    delete[] m_pIHwDelta; delete[] m_pHOwDelta;
}

/* --- Fr die Matrix-Attribute Speicher bereitstellen ------------------- */
void simpleNet::MemAlloc(const size_t dimI, const size_t dimH, const size_t dimO){
    m_amountI = dimI; m_amountH = dimH; m_amountO = dimO;
    
    m_pIHweights = new float*[m_amountI+1];
    for(size_t i=0; i<=m_amountI; i++)
        m_pIHweights[i] = new float[m_amountH];
    m_pHOweights = new float*[m_amountH+1];
    for(size_t i=0; i<=m_amountH; i++)
        m_pHOweights[i] = new float[m_amountO];
    m_pIHwDelta = new float*[m_amountI+1];
    for(size_t i=0; i<=m_amountI; i++)
        m_pIHwDelta[i] = new float[m_amountH];
    m_pHOwDelta = new float*[m_amountH+1];
    for(size_t i=0; i<=m_amountH; i++)
        m_pHOwDelta[i] = new float[m_amountO];
    
    m_pPrefO=new float[m_amountO];
    m_pInI=new float[m_amountI];
    m_pInH=new float[m_amountH];
    m_pInO=new float[m_amountO];
    m_pOutI=new float[m_amountI+1];
    m_pOutH=new float[m_amountH+1];
    m_pOutO=new float[m_amountO];
    m_pErrH=new float[m_amountH+1];
    m_pErrO=new float[m_amountO];
    m_pDiffO=new float[m_amountO];
    // Matrizen auf Startwert 0.0 setzen
    for(size_t i=0; i<m_amountO; i++)
    	m_pDiffO[i]=0.0;
    for(size_t i=0; i<=m_amountI; i++)
        for(size_t j=0; j<m_amountH; j++)
            m_pIHwDelta[i][j]=0.0;
    for(size_t i=0; i<=m_amountH; i++)
        for(size_t j=0; j<m_amountO; j++)
            m_pHOwDelta[i][j]=0.0;
            
    m_learnRate = 1.2;            // Standard-Lernrate setzen
    m_impuls    = 0.3;            // Standard-Impulsfaktor setzen
    m_errTol    = 0.1;            // Standard-Fehlertoleranz setzen
}    

/* --- Matrizen fr einen Vorwrtsdurchlauf auf 0.0 setzen --------------- */
void simpleNet::SetZero(){
    for(size_t i=0; i<m_amountH; i++){
        m_pInH[i]  = 0.0;
        m_pErrH[i] = 0.0;
    }
    for(size_t i=0; i<m_amountO; i++)
        m_pInO[i]  = 0.0;
}

/* --- Mit Backpropagation rckwrts laufen und Gewichte anpassen -------- */
void simpleNet::Backprop(){
    // zuerst die Fehlervektoren des letzten Trainingslaufs berechnen
    for(size_t i=0; i<m_amountO; i++)
        m_pErrO[i] = m_pDiffO[i] * m_pOutO[i] * (1-m_pOutO[i]);
    for(size_t i=0; i<=m_amountH; i++){
        for(size_t j=0; j<m_amountO; j++)
            m_pErrH[i] += m_pErrO[j] * m_pHOweights[i][j];
        m_pErrH[i] *= m_pOutH[i] * (1-m_pOutH[i]);
    }
    // Backpropagation, wie er implementiert ist in Robert Callan [2003]
    // Gewichtsmatrix zwischen verborgener und Ausgabe-Schicht anpassen
    for(size_t i=0; i<=m_amountH; i++)
        for(size_t j=0; j<m_amountO; j++){
            m_pHOwDelta[i][j]  = m_learnRate*m_pErrO[j]*m_pOutH[i] + m_pHOwDelta[i][j]*m_impuls;
            m_pHOweights[i][j] = m_pHOweights[i][j] + m_pHOwDelta[i][j];
        }
    // Gewichtsmatrix zwischen Eingabe- und verborgener Schicht anpassen
    for(size_t i=0; i<=m_amountI; i++)
        for(size_t j=0; j<m_amountH; j++){
		    m_pIHwDelta[i][j]  = m_learnRate*m_pErrH[j]*m_pOutI[i] + m_pIHwDelta[i][j]*m_impuls;
		    m_pIHweights[i][j] = m_pIHweights[i][j] + m_pIHwDelta[i][j];
        }    
}

/* --- Netz mit seinen Dims und Gewichten in Datei speichern ------------- */
void simpleNet::SaveWeightsToFile(const string& filename) throw(ANN_ERROR::file_exception){
	ofstream out(filename.c_str());
    if(!out)
        throw ANN_ERROR::file_exception("Error opening file");
    // Ausnahme werfen bei einem Dateifehler
    out.exceptions(ios_base::failbit|ios_base::badbit|ios_base::eofbit);	
    // Netzdimensionen schreiben
    try{
   	    out << m_amountI << endl;
   	    out << m_amountH << endl;
   	    out << m_amountO << endl;
   	}
   	catch(ios_base::failure){
   	    out.close();
        throw ANN_ERROR::file_exception("Error writing weight matrix dimensions");
    }
    // Gewichtsmatrix 1 schreiben
    for(size_t i=0; i<=m_amountI; i++)
   	    for(size_t j=0; j<m_amountH; j++){
    	    try{
    	        out << m_pIHweights[i][j] << endl;
	        }    
	        catch(ios_base::failure){
                out.close();
	            throw ANN_ERROR::file_exception("Error writing input-hidden weight matrix");
            }
        }
  	// Gewichtsmatrix 2 schreiben
    for(size_t i=0; i<=m_amountH; i++)
  	    for(size_t j=0; j<m_amountO; j++){
 		    try{
      			out << m_pHOweights[i][j] << endl;
			}
			catch(ios_base::failure){
			    out.close();
                throw ANN_ERROR::file_exception("Error writing hidden-output weight matrix");
			}    
   		}
     out.close();
}    

/* --- Rckgabe der Fehlersumme des letzten Trainingslaufs --------------- */
float simpleNet::GetE(){
    float quadsum = 0;
    for(size_t i=0; i<m_amountO; i++)
        quadsum += m_pDiffO[i] * m_pDiffO[i];
    return quadsum;
}

/* --- Vorwrtsdurchlauf durch das Netz, dessen Ausgabe zurckgeben ------ */
float *simpleNet::Run(const float* pInput){
    SetZero();
    for(size_t i=0; i<m_amountI; i++)            // Spezialfall hier: AusgabeI = EingabeI
        m_pOutI[i] = pInput[i];
    m_pOutI[m_amountI] = 1.0;                    // "Shift-Neuron" ist 1.0
    for(size_t i=0; i<m_amountH; i++)            // Eingabe der verborgenen Schicht berechnen
        for(size_t j=0; j<=m_amountI; j++)
            m_pInH[i] += m_pIHweights[j][i] * m_pOutI[j];
    for(size_t i=0; i<m_amountH; i++)            // Ausgabe des verborgenen Schicht (sigmoide Fkt.)
        m_pOutH[i] = 1 / (1 + exp(-m_pInH[i]));
    m_pOutH[m_amountH] = 1;                      // "Shift-Neuron" ist 1.0
    for(size_t i=0; i<m_amountO; i++)            // Eingabe der Ausgabeschicht berechnen
        for(size_t j=0; j<=m_amountH; j++)
            m_pInO[i] += m_pHOweights[j][i] * m_pOutH[j];
    for(size_t i=0; i<m_amountO; i++)            // Ausgabe der Ausgabeschicht mit sigmoider Fkt.
        m_pOutO[i] = 1 / (1 + exp(-m_pInO[i]));
    return m_pOutO;
}    

/* --- Netz trainieren: erst vorwrts durchlaufen, dann Backpropagation -- */
bool simpleNet::Train(const float* pInput, const float* pOutput){
    bool diffInLimit = true;
    // Erster Schritt: Vorwrtsdurchlauf
    Run(pInput);
    for(size_t i=0; i<m_amountO; i++)            // SOLL-Output abspeichern
        m_pPrefO[i] = pOutput[i];
    for(size_t i=0; i<m_amountO; i++)            // Differenzvektor SOLL-IST berechnen
        m_pDiffO[i] = m_pPrefO[i] - m_pOutO[i];
    if(GetE() > m_errTol)                        // wenn Fehler ber Limit -> false zurckgeben
        diffInLimit = false;
    // Zweiter Schritt: Rckwrtsdurchlauf mit Backpropagation
    Backprop();
    return diffInLimit;
}    

/* --- Zurckliefern des Ausgabevektors ---------------------------------- */
float *simpleNet::GetOutput(){ return m_pOutO; }

/* --- Ausgabe-Differenzvektor zurckgeben ------------------------------- */
float *simpleNet::GetODiff(){ return m_pDiffO; }

/* --- Anzahl der Eingabeneuronen zurckgeben ---------------------------- */
size_t simpleNet::GetInputNeurons(){ return m_amountI; }

/* --- Anzahl der verborgenen Neuronen zurckgeben ----------------------- */
size_t simpleNet::GetHiddenNeurons(){ return m_amountH; }

/* --- Anzahl der Ausgabeneuronen zurckgeben ---------------------------- */
size_t simpleNet::GetOutputNeurons(){ return m_amountO; }

/* --- Impulsfaktor zurckgeben ------------------------------------------ */
float simpleNet::GetImpuls(){ return m_impuls; }

/* --- Lernrate zurckgeben ---------------------------------------------- */
float simpleNet::GetLearnRate(){ return m_learnRate; }

/* --- Fehlertoleranz zurckgeben ---------------------------------------- */
void simpleNet::SetErrTolerance(const float tol){ m_errTol = tol; }

/* --- Lernrate setzen --------------------------------------------------- */
void simpleNet::SetLearnRate(const float rate){ m_learnRate = rate; }

/* --- Impulsfaktor setzen ----------------------------------------------- */
void simpleNet::SetImpuls(const float imp){ m_impuls = imp; }
