#ifndef _SIMPLENET_H 
#define _SIMPLENET_H
/***************************************************************************
 *   Copyright (C) 2005 by Matthias Jauernig, Michael Lahl                 *
 *   rtc@linux-related.de, DJ-Mayday@web.de                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

/** \file simpleNet.h
 *  \brief Diese Datei definiert die Neuronale-Netz-Klasse simpleNet.
 *
 * Diese Datei dient zur Definition von simpleNet, einer Klasse, welche
 * Funktionalitt zum Anlernen und Durchlaufen eines Neuronalen Netzes
 * bereitstellt.
 * $Id: simpleNet.h,v 1.8 2005/05/10 14:48:27 djmayday Exp $
 */

/** \brief ANN_ERROR definiert die Exception file_exception.
 *
 * Die Ausnahme ANN_ERROR::file_exception wird geworfen, wenn beim Laden
 * bzw. Speichern des Neuronalen Netzes aus einer bzw. in eine Datei etwas
 * schief gehen sollte.
 */
namespace ANN_ERROR{
    /** \brief Struktur file_exception zur Ausnahmebehandlung.
     *
     * Ein Objekt der struct file_exception wird geworfen, wenn beim Laden
     * bzw. Speichern des Neuronalen Netzes aus einer bzw. in eine Datei
     * etwas schief gehen sollte.
     */
    struct file_exception{
        /**
         * p stellt den String dar, der die Ausnahme charakterisieren soll.
         */
        std::string p;
        
        /**
         * Konstruktor fr ein file_exception-Objekt.\n
         * Dieser setzt das Attribut p auf den bergebenen String q, um der
         * Ausnahme eine individuelle Meldung zuzuordnen.
         * \param q String, welcher den Grund der Ausnahme beschreibt.
 		 */
		file_exception(const std::string& q);
	};       
}

/** \brief Einfache Neuronale-Netz-Klasse.
 *  
 * simpleNet implementiert ein einfaches Neuronales Netz. Dieses Netz hat
 * unterschiedliche Eigenschaften. Zum einen ist es vorwrtsgerichtet (feed forward),
 * weiterhin ist es ein 3-Schichten-Netz mit Eingabeschicht, verborgener Schicht und
 * Ausgabeschicht. Zum dritten findet beim Lernprozess ein berwachtes Lernen mittels
 * Backpropagation-Algorithmus statt.\n
 * simpleNet kann effektiv zum Anlernen verschiedener Funktionen (Menge von Tupeln)
 * benutzt werden, indem man die Methode Train() aufruft. Nach der Trainingsphase
 * kann schnell fr jeglichen Input der entsprechende Output durch Aufruf von Run()
 * erhalten werden, was die grbste Funktionalitt des Netzes auf diese beiden
 * Methoden einschrnkt (was nicht heien soll, das dies schon alles war ;)).
 */
class simpleNet{
    private:
        float **m_pIHweights, **m_pHOweights;
        float **m_pIHwDelta, **m_pHOwDelta;
        float *m_pInI, *m_pInH, *m_pInO;
        float *m_pOutI, *m_pOutH, *m_pOutO;
        float *m_pErrH, *m_pErrO, *m_pPrefO, *m_pDiffO;
        size_t m_amountI, m_amountH, m_amountO;
        float m_errTol, m_learnRate, m_impuls;
        void SetZero();
        void MemAlloc(const size_t dimI, const size_t dimH, const size_t dimO);
        void Backprop();
    public:
        /** 
         * Dieser Konstruktor erzeugt das Neuronale Netz mit den bergebenen Dimensionen, die
         * fr die Anzahl der Neuronen in der jeweiligen Schicht stehen. Dazu wird fr alle
         * Matrizen und Vektoren Speicher allokiert und die Elemente der Gewichtsmatrizen mit
         * zuflligen Werten zwischen [-0.3 .. +0.3] belegt.
         * \param dimI Anzahl der Input-Neuronen.
         * \param dimH Anzahl der verborgenen Neuronen.
         * \param dimO Anzahl der Output-Neuronen.
         */
        simpleNet(const size_t dimI, const size_t dimH, const size_t dimO);
        
        /**
         * Dieser Konstruktor erzeugt das Neuronale Netz mit den bergebenen Dimensionen, die
         * fr die Anzahl der Neuronen in der jeweiligen Schicht stehen. Dazu wird fr alle
         * Matrizen und Vektoren Speicher allokiert. Weiterhin werden die Gewichtsmatrizen mit
         * den Werten aus den bergebenen Matrizen ihW und hoW belegt.
         * \param dimI Anzahl der Input-Neuronen.
         * \param dimH Anzahl der verborgenen Neuronen.
         * \param dimO Anzahl der Output-Neuronen.
         * \param ihW  Gewichtsmatrix zwischen Eingabe- und verborgener Schicht. Ihre Elemente
         *             mssen zwischen -1.0 und +1.0 liegen.
         * \param hoW  Gewichtsmatrix zwischen verborgener und Ausgabe-Schicht. Ihre Elemente
         *             mssen zwischen -1.0 und +1.0 liegen.
         */
        simpleNet(const size_t dimI, const size_t dimH, const size_t dimO,
            const float **ihW, const float **hoW);
        
        /**
         * Der dritte Konstruktor erzeugt das Neuronale Netz, indem er die Dimensionen und
         * die Gewichtsmatrizen aus der Datei filename einliest. Sollte das Einlesen
         * schief gehen, z.B. weil es sich um keine korrekte Datei fr das Neuronale Netz
         * handelt, so wird eine ANN_ERROR::file_exception geworfen.
         * \param filename Name der Datei, aus der das Neuronale Netz geladen werden soll.
         */
        simpleNet(const std::string& filename) throw(ANN_ERROR::file_exception);
        
        /**
         * Destruktor des Netzes - lscht bereitgestellten Speicherplatz.
         */
        ~simpleNet();
        
        /**
         * Mit dieser Methode kann das aktuelle Neuronale Netz in der Datei filename
         * abgespeichert werden, um es spter ganz einfach laden zu knnen und sich
         * so den langwierigen Prozess des Anlernens zu ersparen.\n
         * Beim Speichern werden die Dimensionen des Netzes und danach die beiden
         * Gewichtsmatrizen in die Datei geschrieben.\n
         * Sollte das Speichern schief gehen, so wird eine ANN_ERROR::file_exception geworfen.
         * \param filename Name der Datei, in die das Netz gespeichert werden soll.
         */
        void SaveWeightsToFile(const std::string& filename) throw(ANN_ERROR::file_exception);
        
        /**
         * Diese Methode liefert den quadratischen Fehler, d.h. die quadratische Abweichung
         * zwischen IST- und SOLL-Wert, an der Ausgangsschicht zurck, der durch den letzten
         * Aufruf von train() erreicht wurde.
         * \return Quadratischer Fehler, der beim letzten Train()-Aufruf gemacht wurde.
         */
        float GetE(void);
        
        /**
         * Es wird die Ausgabe des letzten Run()-Aufrufs zurck gegeben.
         * \return Ausgabevektor, vom letzten Run()-Befehl verursacht.
         */
        float *GetOutput(void);
        
        /**
         * Rckgabe des Vektors mit der Differenz zwischen IST- und SOLL-Wert an der
         * Ausgabeschicht des Neuronalen Netzes, welcher vom letzten Train()-Aufruf
         * verursacht wurde.
         * \return Differenzvektor zwischen IST- und SOLL-Ausgabe des letzten Train()-Aufrufs.
         */
        float *GetODiff(void);
        
        /**
         * Gibt die Anzahl der Eingabeneuronen zurck.
         * \return Anzahl der Eingabeneuronen.
         */
        size_t GetInputNeurons(void);

        /**
         * Gibt die Anzahl der verborgenen Neuronen zurck.
         * \return Anzahl der verborgenen Neuronen.
         */
        size_t GetHiddenNeurons(void);

        /**
         * Gibt die Anzahl der Ausgabeneuronen zurck.
         * \return Anzahl der Ausgabeneuronen.
         */
        size_t GetOutputNeurons(void);
        
        /**
         * Gibt den aktuellen Impulsfaktor zurck.
         * \return Aktuell gesetzter Impulsfaktor.
         */
        float GetImpuls(void);

        /**
         * Gibt die aktuelle Lernrate zurck.
         * \return Aktuell gesetzte Lernrate.
         */        
        float GetLearnRate(void);
        
        /**
         * Setze die Fehlertoleranz auf den bergebenen Wert tol.\n
         * Die Fehlertoleranz gibt den Wert an, ab welchem Train()
         * true zurckliefert. Dies ist genau dann der Fall, wenn der
         * Betrag des Differenzvektors m_pDiffO kleiner als die Fehlertoleranz ist.\n
         * Der Default-Wert, welcher beim Erzeugen des Netzes gesetzt wird, ist 0.1.
         * \param tol Wert, auf welchen die Fehlertoleranz gesetzt werden soll.
         */
        void SetErrTolerance(const float tol);
        
        /**
         * Setze die Lernrate auf den bergebenen Wert rate.\n
         * Die Lernrate gibt an, wie schnell das Netz einen Input lernt. Sollte sie
         * jedoch zu gro gewhlt werden, so wird das Fehler-Minimum, welches beim
         * Lernen erreicht werden soll, berschritten und das Netz kann nicht mehr
         * korrekt angelernt werden. Der Wert ist abhngig vom Impulsfaktor und drfte
         * mit 1.2 fr die meisten Probleme gut gewhlt sein.\n
         * Der Default-Wert, welcher beim Erzeugen des Netzes gesetzt wird, ist 1.2.
         * \param rate Wert, auf welchen die Lernrate gesetzt werden soll.
         */
        void SetLearnRate(const float rate);
        
        /**
         * Setze den Impulsfaktor auf den bergebenen Wert rate.\n
         * Der Impulsfaktor gibt an, mit welchem Faktor multipliziert die Gewichtsnderung
         * des letzten Train()-Aufrufs zur aktuellen Gewichtsnderung addiert werden soll.
         * Daher darf dieser Term nicht grer als 1.0 sein. Wird der Impulsfaktor zu hoch
         * gewhlt, so wird wie bei einer zu hohen Lernrate das Fehler-Minimum brschritten
         * und das Netz kann nicht mehr korrekt angelernt werden. Der Wert ist vom gewhlten
         * Lernfaktor abhngig und sollte mit 0.3 fr die meisten Probleme gut gewhlt sein.\n
         * Der Default-Wert, welcher beim Erzeugen des Netzes gesetzt wird, ist 0.3.
         * \param imp Wert, auf welchen der Impulsfaktor gesetzt werden soll.
         */
        void SetImpuls(const float imp);
        
        /**
         * Mit dem bergebenen Input (welcher einen float-Wert fr jedes Eingabeneuron enthlt)
         * wird das Netz in der aktuellen Konfiguration einmal vorwrts durchlaufen.
         * Der dabei ermittelte Output (mit einem Wert fr jedes Output-Neuron) wird als Vektor
         * zurckgegeben und kann dann vom Aufrufer ausgewertet werden. Sollte das Netz bereits
         * trainiert worden sein, kann der Aufrufer den vom Netz zum bergebenen Input
         * gefundenen Output hiermit ermitteln.
         * \param pInput Vektor mit dem zu verarbeitenden Input. Dieser muss von der Gre der
         *               (Anzahl der) Eingabeneuronen sein.
         * \return Vektor mit dem vom Netz ermittelten Output, welcher genau soviele Elemente
         *         hat wie es Neuronen in der Ausgabeschicht gibt.
         */
        float *Run(const float* pInput);
        
        /**
         * Mit dieser Methode wird genau ein Trainingslauf durch das Neuronale Netz mit der
         * bergebenen Input-/Output-Kombination durchgefhrt. D.h. das Netz wird daraufhin
         * angelernt, dass es bei einem Run() mit pInput mglichst den Vektor pOutput
         * zurckliefern sollte, eine gewisse Fehlertoleranz eingeschlossen.\n
         * Zunchst ermittelt Train() den Fehlerdifferenz-Vektor, welcher momentan gemacht
         * werden wrde bei einem Durchlauf mittels Run(pInput). Dann wird der
         * Backpropagation-Algorithmus angewendet, um die Gewichte in Hinblick auf einen
         * mglichst geringen Fehler anzupassen (mit der definierten Lernrate und dem
         * gesetzten Impulsfaktor). Sollte der Betrag des Fehler-Vektors unter der
         * Fehlertoleranz der Netzes liegen, so wird true zurckgegeben.
         * \param pInput  Eingabevektor, welcher dem Netz vorgelegt werden soll.
         * \param pOutput Ausgabevektor, der mit pInput erreicht werden soll.
         * \return        true, wenn der Fehler der Ausgabeschicht unterhalb der definierten
         *                Fehlertoleranz liegt.\n
         *                false, wenn die Fehlertoleranz berschritten wird.
         */
        bool Train(const float* pInput, const float* pOutput);
};

#endif
